/*
	Name: usb.c
	Date: 1.Mar.2013
	Encoding: UTF-8
	Author: Michal Riša <xrisam01@stud.fit.vutbr.cz>
	Licence:
		Copyright (C) 2013 Michal Riša <mh74s74@gmail.com>

		This file is part of bp_xrisam01.

		bp_xrisam01 is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License version 3 as
		published by the Free Software Foundation.

		bp_xrisam01 is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
		GNU General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with bp_xrisam01. If not, see <http://www.gnu.org/licenses/>.
*/

#include <hidef.h>
#include "derivative.h"
#include "usb.h"
#include "usb_drv.h"

extern const uint8_t cfg0[];

//buffer endpointu 1
uint8_t ep1_buf[4] @0x18A0 = {0};

//------------------------------------------------------------------------------
//Prototypy
static void adc_init(void);
static uint8_t move_px(uint8_t diff);
static void kb_init(void);
static void key_fsm(uint8_t key, uint8_t event);
static uint8_t is_key_down(uint8_t key);
static void ep1_data(void);
static void rtc_init(void);
void mouse_cfg(uint8_t event);
static void mouse_send_data(void);

//------------------------------------------------------------------------------
//Posielanie dat USB mysi

//buffer dat na poslanie (nie buffer endpointu)
uint8_t mouse_data[4] = {0}; //kludovy stav

//indexy do buffra
enum mouse_buf_idx {
	MOUSE_BUTTONS = 0,
	MOUSE_X,
	MOUSE_Y,
	MOUSE_SCROLL,
};
//------------------------------------------------------------------------------
//ADC

enum { /* kanaly ADC pripojene na akcelerometer */
	ADC_Y = 0,
	ADC_X = 1,
};

/*
Funkcia: adc_init
Popis: Zapne ADC v rezime opakovanych konverzii.
*/
static void adc_init(void) {
	/* clock, long sample time, 8 bit result, low power config,
	   bus clock / 2 / 8 = 1.5MHz */
	ADCCFG = 0b11110001;
	//conversion trigger
	ADCSC2 = 0b00000000;
	//long trva 40 cyklov ADC, short 20 cyklov ADC
	//continuous conversions, interrupt, input channel 0
	//ADP0 = Y, ADP1 = X
	ADCSC1 = 0b01100000;
}

/*
Funkcia: move_px
Parametre: Hodnota naklonu z ADC prevodnika.
Navratova hodnota: Posun kurzoru v pixeloch.
Popis: Prevedie naklon na posun kurzoru s priblizne logaritmickym meritkom.
*/
static uint8_t move_px(uint8_t diff) {
	//10*log(diff/3)
	if(diff < 4)
		return 0;
	if(diff < 9)
		return 1;
	if(diff < 13)
		return 2;
	if(diff < 17)
		return 3;
	if(diff < 21)
		return 4;
	if(diff < 25)
		return 5;
/*	if(diff < 12)
		return 6;
	if(diff < 16)
		return 7;
	if(diff < 19)
		return 8;
	return 9;*/
	return 6;
}

#define MOUSE_MOVING (ADCRL < 56 || ADCRL > 96)

/*
Funkcia: adc_isr
Popis: Obsluzna funkcia ADC prerusenia.
*/
__interrupt VectorNumber_Vadc void adc_isr(void) {
	//ADC kanal s momentalnym vysledkom
	static uint8_t channel = 0;
	uint8_t dst_idx;
	//momentalny vysledok prevodu
	uint8_t value = ADCRL;

	if(ADCRL < 72) {
		//10*log((76-ADCRL)/3)
		value = -move_px(76 - ADCRL);
	} else if(ADCRL > 80) {
		//10*log((ADCRL-76)/3)
		value = move_px(ADCRL - 76);
	} else
		value = 0;

	//podla kanalu vyberie smer pohybu
	dst_idx = channel ? MOUSE_Y : MOUSE_X;

	//nabuduce prevadza druhy kanal
	channel = channel ? 0 : 3;

	if(mouse_data[dst_idx] == 0)
		//neprepise stav
		mouse_data[dst_idx] = value;

	//buduci prevod
	ADCSC1 = 0b01100000 | channel;
}

#define ADC_SUSPEND ADCSC1_AIEN = 0
#define ADC_WAKEUP  ADCSC1_AIEN = 1
//------------------------------------------------------------------------------
//Tlacitka
#define KEY_LEFT_VALUE        PTGD_PTGD2
#define KEY_RIGHT_VALUE       PTGD_PTGD3
#define KEY_SCROLL_UP_VALUE   PTGD_PTGD0
#define KEY_SCROLL_DOWN_VALUE PTGD_PTGD1

enum key_state {
	KS_UP = 0, //nestlacene tlacitko
	KS_INT, //tlacitko po prichode falling edge prerusenia
	KS_DOWN //tlacitko stlacene
};

enum key_mask {
	KEY_LEFT_MASK        = 0b00000100,
	KEY_RIGHT_MASK       = 0b00001000,
	KEY_SCROLL_UP_MASK   = 0b00000001,
	KEY_SCROLL_DOWN_MASK = 0b00000010,
};

enum key_number {
	KEY_LEFT = 0,
	KEY_RIGHT = 1,
	KEY_SCROLL_UP = 2,
	KEY_SCROLL_DOWN = 3,
};

enum key_event {
	KE_KBI,
	KE_TIMER
};

uint8_t key_state[4] = {0}; //KS_UP - vsetky tlacitka nestlacene

static const uint8_t key_mask[4] = {
	KEY_LEFT_MASK,
	KEY_RIGHT_MASK,
	KEY_SCROLL_UP_MASK,
	KEY_SCROLL_DOWN_MASK,
};

/*
Funkcia: kb_init
Popis: Zapne KBI modul s aktivovanymi vyvodmi 7 6 1 0, citlivostou na zostupnu
       hranu a povoli KBI prerusenie.
*/
static void kb_init(void) {
	//pull-up odpory
	PTGPE = 0b00001111;

	//aktivacia KBI vyvodov 7, 6, 1, 0
	KBIPE = 0b11000011;

	//citlivost na zostupnu hranu
	KBIES = 0;

	//povolenie prerusenia
	KBISC_KBIE = 1;
}

/*
Funkcia: key_fsm
Parametre:
	key: enum key_number: cislo tlacitka.
	event: enum key_event: udalost.
Popis: Konecny automat sledujuci stisk a podrzanie tlacitok. Ignoruje zakmity.
*/
static void key_fsm(uint8_t key, uint8_t event) {
	switch(event) {
		case KE_KBI:
			//KBI udalost: klesajuca hrana na KBI pine
			if(key_state[key] == KS_UP &&
			  (PTGD & key_mask[key]) == 0)
				//prvy stisk tlacitka
				key_state[key] = KS_INT;
			break;
		case KE_TIMER:
			//Vyprsal RTC
			if(key_state[key] == KS_INT) {
				if((PTGD & key_mask[key]) == 0)
					key_state[key] = KS_DOWN;
				else
					key_state[key] = KS_UP;
			} else if(key_state[key] == KS_DOWN) {
				if((PTGD & key_mask[key]) == 1)
					key_state[key] = KS_UP;
			}
			break;
	}
}

/*
Funkcia: is_key_down
Parametre:
	key: enum key_number cislo tlaticka.
Navratova hodnota:
	1: tlacitko je stlacene, inak je uvolnene.
Popis: ^.
*/
static uint8_t is_key_down(uint8_t key) {
	return key_state[key] == KS_DOWN && (PTGD & key_mask[key]) == 0;
}

/*
Funkcia: kbi_isr
Popis: Obsluzna funkcia KBI prerusenia.
*/
__interrupt VectorNumber_Vkeyboard void kbi_isr(void) {
	if(PTGD_PTGD0 == 0) //key_scroll_up
		key_fsm(KEY_SCROLL_UP, KE_KBI);
	if(PTGD_PTGD1 == 0) //key_scroll_down
		key_fsm(KEY_SCROLL_DOWN, KE_KBI);
	if(PTGD_PTGD2 == 0) //key_left
		key_fsm(KEY_LEFT, KE_KBI);
	if(PTGD_PTGD3 == 0) //key_right
		key_fsm(KEY_RIGHT, KE_KBI);
	KBISC_KBACK = 1; /* odstrani priznak prerusenia */
}
//------------------------------------------------------------------------------
//Hodiny realneho casu

/*
Funkcia: rtc_init
Popis: Zapne hodiny realneho casu s periodou 4 milisekundy.
*/
static void rtc_init(void) {
	RTCMOD = 0x00;
	RTCSC = 0b00011101;
}

#define RTC_SUSPEND RTCSC_RTIE = 0
#define RTC_WAKEUP  RTCSC_RTIE = 1

/*
Funkcia: rtc_isr
Popis: Obsluzna funkcia RTC prerusenia. Prepocita stav tlacitok, posklada nove
       honoty tlacitok a da ich modulu pre posielanie dat.
*/
__interrupt VectorNumber_Vrtc void rtc_isr(void) {
	uint8_t tmp;

	RTCSC |= 0x80;
	if(RTCSC_RTIE == 0)
		return;

	//prepocita stav tlacitok
	key_fsm(KEY_LEFT,        KE_TIMER);
	key_fsm(KEY_RIGHT,       KE_TIMER);
	key_fsm(KEY_SCROLL_UP,   KE_TIMER);
	key_fsm(KEY_SCROLL_DOWN, KE_TIMER);

	//left, right tlacitka
	tmp = 0;
	if(is_key_down(KEY_LEFT))
		tmp |= 0x1;
	if(is_key_down(KEY_RIGHT))
		tmp |= 0x2;
	//premaze predosly stav
	mouse_data[MOUSE_BUTTONS] = tmp;

	//scroll up, down tlacitka
	if(is_key_down(KEY_SCROLL_UP))
		tmp = 0x01;
	else if(is_key_down(KEY_SCROLL_DOWN))
		tmp = 0xFF;
	else
		tmp = 0;
	if(tmp != 0)
		//nepremaze predosly stav
		mouse_data[MOUSE_SCROLL] = tmp;

	//posle data
	mouse_send_data();
}
//------------------------------------------------------------------------------
//spracovanie specifickych prikazov (HID class specific requests)

/*
Funkcia: usb_OnClassReq
Navratova hodnota:  0 = chyba prikazu, inak ok.
Popis: Obsluzna funkcia specifickych prikazov USB mysi.
*/
uint8_t usb_OnClassReq(void) {
	//HID report
	static  const uint8_t hid_report[] = {
		0x05, 0x01,
		0x09, 0x02,
		0xa1, 0x01,
		0x09, 0x01,
		0xa1, 0x00,
		0x05, 0x09,
		0x19, 0x01,
		0x29, 0x05,
		0x15, 0x00,
		0x25, 0x01,
		0x95, 0x05,
		0x75, 0x01,
		0x81, 0x02,
		0x95, 0x01,
		0x75, 0x03,
		0x81, 0x01,
		0x05, 0x01,
		0x09, 0x30,
		0x09, 0x31,
		0x09, 0x38,
		0x15, 0x81,
		0x25, 0x7f,
		0x75, 0x08,
		0x95, 0x03,
		0x81, 0x06,
		0xc0,
		0xc0,
	};

	enum hid_requests {
		GET_REPORT   = 1,  //povinny
		GET_IDLE     = 2,
		GET_PROTOCOL = 3,  //vyzadovany pre fungovanie v biose
		SET_REPORT   = 9,
		SET_IDLE     = 10,
		SET_PROTOCOL = 11, //vyzadovany pre fungovanie v biose
	};

	enum hid_descriptors {
		HID_DESCRIPTOR = 33,
		REPORT_DESCRIPTOR = 34,
		PHYSICAL_DESCRIPTOR = 35,
	};

	enum hid_reports {
		REPORT_INPUT = 1,
		REPORT_OUTPUT = 2,
		REPORT_FEATURE = 3,
	};

	if(ep0out_setup.bytes.bRequest == GET_DESCRIPTOR) {
		/* setup:
			bRequest	GET_DESCRIPTOR
			wValue		[1] = typ, [0] = index
			wIndex		
			wLength		pozadovany pocet bajtov na prenesenie */

		volatile uint8_t type = ep0out_setup.bytes.wValue[1];
		switch(type) {
			case HID_DESCRIPTOR:
				/*	18: hid descriptor je vnoreny v cfg0 deskriptore,
					pred nim su configuration a interface deskriptory zaberajuce
					9 (CONFIGURATION_DESCRIPTOR_BYTES) +
					9 (INTERFACE_DESCRIPTOR_BYTES) bajtov => 18 */
				ctrl.data = (uint8_t*)cfg0 + 18; //cfg0 od 19. bajtu
				ctrl.bytes = *ctrl.data; //bLength
				return 1;
			case REPORT_DESCRIPTOR:
				ctrl.data = (uint8_t*)hid_report;
				ctrl.bytes = sizeof hid_report;
				return 1;
		}
		return 0;
	}

	if(ep0out_setup.bytes.bRequest == GET_INTERFACE) {
		//aktivne je rozhranie 0
		ctrl.data = NULL;
		ctrl.bytes = 1;
		ep0in_buf[0] = 0;
		return 1;
	}

	if(ep0out_setup.bmRequestType.Type != TYPE_CLASS)
		//standardny prikaz tu nema co robit
		return 0;

	switch(ep0out_setup.bytes.bRequest) {
		case GET_REPORT:
			/* setup:
				bRequest	GET_REPORT
				wValue		[1] = typ, [0] = index
				wIndex		Interface
				wLength		pozadovany pocet bajtov na prenesenie */
			if(ep0out_setup.bytes.wValue[0] != REPORT_INPUT /* typ reportu */ ||
				ep0out_setup.bytes.wValue[1] != 0 /* index reportu */)
				return 0;
			ctrl.data = (uint8_t*)hid_report;
			ctrl.bytes = sizeof hid_report;
			return 1;
/*		case GET_PROTOCOL:
			return 1;
		case SET_PROTOCOL:
			return 1;*/
	}
	return 0;
}
//------------------------------------------------------------------------------
//Obsluzne funkcie aplikacie
/*
Funkcia: mouse_cfg
Parametre:
	event: udalost konfiguracie, 1 z EVENT_:
		1) ACTIVATE: aktivacia tejto konfiguracie
		2) DEACTIVATE: deaktivacia tejto konfiguracie
		3) SUSPEND: suspend zariadenia
		4) RESUME: resume zariadenia
Popis: Obsluzna funkcia udalosti tejto konfiguracie.
*/
void mouse_cfg(uint8_t event) {
	switch(event) {
		case EVENT_ACTIVATE:
			//aktivacia konfiguracie

			//obsluzna funkcia endpointu 1
			ep_hdrs[1] = ep1_data;
			//interrupt in endpoint 1
			bdt.ep1.u.reg = EPF_OWN_MCU | EPF_DATA0;
			bdt.ep1.epadr = EPADR(ep1_buf);
			//zapne endpoint 1
			EPCTL1 = EP_TX | EP_HSK;

			//zapne periferie
			rtc_init();
			kb_init();
			adc_init();
			break;

		case EVENT_DEACTIVATE:
			//vypne endpoint
			EPCTL1 = 0;
		case EVENT_SUSPEND:
			//vypne pefirerie
			RTC_SUSPEND;
			ADC_SUSPEND;
			break;

		case EVENT_WAKEUP:
			//zapne periferie
			RTC_WAKEUP;
			ADC_WAKEUP;
			break;
	}
}

/*
Funkcia: usb_OnSof
Popis: Obsluzna funkcia sycnhronizacie isochronnych endpointov.
*/
void usb_OnSof(void) {

}

/*
Funkcia: usb_OnStallSent
Parametre:
	ep: cislo endpointu ktory poslal STALL handshake.
Popis: Obsluzna funkcia poslania STALL handshake.
*/
void usb_OnStallSent(uint8_t ep) {

}

/*
Funkcia: usb_OnError
Popis: Obsluzna funkcia chybovych stavov endpointov 1 az 7.
*/
void usb_OnError(void) {

}

/*
Funkcia: usb_CheckRemoteWakeup
Navratova hodnota: 0 = nenastala podmienka pre remote wakeup, inak nastala
Popis: Obsluzna funkcia ktorou moze aplikacia povedat ovladacu self-powered
       zariadenia so zapnutym remote wakeup ktore je v suspended stave ze
       ma poslat RESUME signalizaciu hostovi.
*/
uint8_t usb_CheckRemoteWakeup(void) {
	return 0; //ziadny remote wakeup
}
//------------------------------------------------------------------------------
//Posielanie dat USB mysi

/*
Funkcia: ep1_data
Popis: Obsluzna funkcia interrupt in endpointu.
*/
static void ep1_data(void) {
	/* vysielajuci endpoint neguje data toggle bit po spravnom odoslani
		transakcie */
	bdt.ep1.u.bits.DATA = !bdt.ep1.u.bits.DATA;

	//posle data ak nejake ma
	mouse_send_data();
}

#define DIFF(a, b)\
	((a) - (b)) < 0 ? ((b) - (a)) : ((a) - (b))

/*
Funkcia: mouse_send_data
Popis: Posiela data USB mysi. Data neposle ak je endpoint vo vlastnictve SIE
       alebo sa nic nedeje. Level-triggered hodnoty (scroll, pohyb) posiela
       periodicky ak su nenulove.
*/
static void mouse_send_data(void) {

	if(bdt.ep1.u.bits.OWN == 1)
		//SIE vlastni endpoint
		return;

	if(mouse_data[MOUSE_BUTTONS] != ep1_buf[MOUSE_BUTTONS] ||
		//zmena stlacenia left/right tlacitka - edge triggered

		mouse_data[MOUSE_SCROLL] != 0 ||
		//stlaceny scroll UP/DOWN - level triggered

		mouse_data[MOUSE_X] != 0 ||
		//X pohyb

		mouse_data[MOUSE_Y] != 0
		//Y pohyb
		) {

	} else
		return;

	ep1_buf[0] = mouse_data[0];
	ep1_buf[1] = mouse_data[1];
	ep1_buf[2] = -mouse_data[2];
	ep1_buf[3] = mouse_data[3];

	//posle data
	bdt.ep1.bc = sizeof ep1_buf;
	if(bdt.ep1.u.bits.DATA)
		bdt.ep1.u.reg = EPF_OWN_SIE | EPF_DTS | EPF_DATA1;
	else
		bdt.ep1.u.reg = EPF_OWN_SIE | EPF_DTS | EPF_DATA0;

	//odstrani priznaky zmien
//	mouse_data[MOUSE_BUTTONS] = 0;
	mouse_data[MOUSE_SCROLL]  = 0;
	mouse_data[MOUSE_X]       = 0;
	mouse_data[MOUSE_Y]       = 0;
}
